import os

import nest_asyncio

nest_asyncio.apply()
# os.environ["KIVY_NO_CONSOLELOG"] = "1"

from importlib import __import__
from kivy.app import App
from kivy.uix.button import Button
from kivy.uix.label import Label
from kivy.uix.floatlayout import FloatLayout
from kivy.config import Config
from kivy.uix.popup import Popup
import kivy.utils

# from .Backend import Backend
# from .helper import resource_path
# from .constants import *
# from .MuseFinder import MuseFinder


Config.set('graphics', 'width', '750')
Config.set('graphics', 'height', '400')
Config.set('input', 'mouse', 'mouse,multitouch_on_demand')
from kivy.uix.image import Image
from kivy.uix.checkbox import CheckBox
from kivy.uix.textinput import TextInput
from kivy.uix.spinner import Spinner
from kivy.graphics import Color, Rectangle

from kivy.resources import resource_add_path
import asyncio
import pkg_resources


from uvicmuse.helper import resource_path
from uvicmuse.constants import *

# Frontend Test Branch
# noinspection PyBroadException,PyUnusedLocal
class UVicMuse(FloatLayout):

    def __init__(self, **kwargs):
        super(UVicMuse, self).__init__(**kwargs)
        self.press_search_txt = "Search for a list of Available Muses on BLE"
        self.btn_color = (204 / 256, 213 / 256, 216 / 256, 1)
        self.txt_color = kivy.utils.get_color_from_hex("#5a636c")
        self.chbx_color = kivy.utils.get_color_from_hex("#0f0f0f")
        # self.txt_color = kivy.utils.get_color_from_hex("#F8B195")
        self.muses = []
        self.sock = None
        self.muse = None
        self.did_connect = False
        self.udp_address = 'localhost'
        self.connected_address = ""
        self.host_address = 'localhost'
        self.current_muse_id = 0

        def draw_background(widget, prop):
            with widget.canvas.before:
                Color(rgba=(240 / 256, 240 / 256, 240 / 256, 1))
                Rectangle(pos=self.pos, size=self.size)

        self.bind(size=draw_background)
        self.bind(pos=draw_background)

        # Create UVic Muse Logo
        DATA_PATH = pkg_resources.resource_filename('uvicmuse', 'docs/')
        self.img = Image(source=os.path.join(DATA_PATH, 'Header.png'), allow_stretch=True)

        # Initiate Labels
        self.status_label = Label(
            text="Press Search to Look For Nearby Muse",
            color=self.txt_color, font_size='16sp', pos_hint={'x': 0.033, 'y': .2*3/2-0.15}, size_hint=(1.0, 1.0), halign="left",
            valign="middle")
        self.status_label.bind(size=self.status_label.setter('text_size'))

        self.filter_title = Label(text="Filters", color=self.txt_color, font_size='18sp', bold=True,
                                  pos_hint={'x': -0.4, 'y': -.015*3/2-0.20})

        self.about_button = Button(text="About Us", size_hint=(.15, .07*3/2), pos_hint={'x': 0.82, 'y': 0.5-0.2*3/2-0.15},
                                   background_color=self.btn_color,
                                   on_release=self.about)

        self.search_button = Button(text="Search", size_hint=(.15, .07*3/2), pos_hint={'x': 0.82, 'y': .5+.1*3/2-0.15},
                                    background_color=self.btn_color, on_press=self.update_status_search,
                                    on_release=self.search)

        # Initiate Buttons and bind press and release to functions
        self.connect_button = Button(text="Connect", size_hint=(.15, .07*3/2), pos_hint={'x': 0.82, 'y': .5+0.0*3/2-0.15},
                                     background_color=self.btn_color,
                                     on_release=self.connect, on_press=self.on_connect_press)
        self.stream_button = Button(text="Start Stream", size_hint=(.15, .07*3/2), pos_hint={'x': 0.82, 'y': 0.5-0.1*3/2-0.15},
                                    background_color=self.btn_color,
                                    on_release=self.stream)

        self.lowpass_label = Label(text="Lowpass", color=self.txt_color, font_size='16sp',
                                   pos_hint={'x': -0.25, 'y': .035*3/2-0.2}, halign='left', valign='middle')
        self.notch_label = Label(text="Notch (60Hz)", color=self.txt_color, font_size='16sp',
                                 pos_hint={'x': -0.25, 'y': -.015*3/2-0.2}, halign='left', valign='middle')
        self.highpass_label = Label(text="Highpass", color=self.txt_color, font_size='16sp',
                                    pos_hint={'x': -0.25, 'y': -.065*3/2-0.2}, halign='left', valign='middle')

        # initiate List with max height
        self.list_box = Spinner(text=self.press_search_txt, values='', size_hint=(0.77, 0.07*3/2),
                                pos_hint={'x': 0.03, 'y': .5+0.1*3/2-0.15}, background_color=self.btn_color)
        self.list_box.dropdown_cls.max_height = self.list_box.height * 1.6

        self.lowpass_checkbox = CheckBox(active=False, size_hint_y=0.02*3/2, size_hint_x=0.02,
                                         pos_hint={'x': 0.4, 'y': 0.5 + 0.025*3/2-0.2}, color=self.chbx_color)
        self.notch_checkbox = CheckBox(active=True, size_hint_y=0.02*3/2, size_hint_x=0.02,
                                       pos_hint={'x': 0.4, 'y': 0.5 - 0.025*3/2-0.2}, color=self.chbx_color)
        self.highpass_checkbox = CheckBox(active=False, size_hint_y=0.02*3/2, size_hint_x=0.02,
                                          pos_hint={'x': 0.4, 'y': 0.5 - 0.075*3/2-0.2}, color=self.chbx_color)
        # Initiate textbox's to enter text
        self.lowpass_text = TextInput(font_size='14sp', pos_hint={"x": 0.50, "y": 0.5 + .015*3/2-0.2}, size_hint=(0.07, 0.05*3/2),
                                      multiline=False, text='30', write_tab=False, halign='center',
                                      background_color=(204 / 256, 213 / 256, 216 / 256, 1))
        self.highpass_text = TextInput(font_size='14sp', pos_hint={"x": 0.50, "y": 0.5 - 0.085*3/2-0.2}, size_hint=(0.07, 0.05*3/2),
                                       multiline=False, text='0.1', write_tab=False, halign='center',
                                       background_color=(204 / 256, 213 / 256, 216 / 256, 1))



        # add widgets that have been initiated to frame
        self.add_widget(self.img)
        self.add_widget(self.about_button)
        self.add_widget(self.search_button)
        self.add_widget(self.connect_button)
        self.add_widget(self.stream_button)
        self.add_widget(self.status_label)
        self.add_widget(self.lowpass_text)
        self.add_widget(self.highpass_text)
        self.add_widget(self.lowpass_label)
        self.add_widget(self.highpass_label)
        self.add_widget(self.lowpass_checkbox)
        self.add_widget(self.highpass_checkbox)
        self.add_widget(self.notch_checkbox)
        self.add_widget(self.notch_label)
        self.add_widget(self.list_box)
        self.add_widget(self.filter_title)

        # Adjust positions of widgets that have been added to the frame
        self.img.pos_hint = {'x': 0, 'y': 0.7}
        self.img.size_hint = (1, 0.3)

        self.notch_checkbox.disabled = True
        self.lowpass_checkbox.disabled = True
        self.highpass_checkbox.disabled = True
        self.highpass_text.disabled = True
        self.lowpass_text.disabled = True
        self.stream_button.disabled = True
        self.connect_button.disabled = True

        from uvicmuse.Backend import Backend

        self.backend = Backend()

    # logic

    def on_connect_press(self, event):
        if self.did_connect:
            pass
        else:
            self.status_label.text = "Connecting to the selected MUSE..."

    # As Search takes 10 seconds, update the status before the process begins
    def update_status_search(self, event):
        self.status_label.text = "Searching for nearby Muses, this may take up to ten seconds..."

    # Function to change state of all components
    def button_state(self, search_state, stream_state, notch_state, lowpass_check_state,
                     lowpass_cut_state, highpass_cut_state, list_box_state):
        self.search_button.disabled = search_state
        self.stream_button.disabled = stream_state
        self.notch_checkbox.disabled = notch_state
        self.lowpass_checkbox.disabled = lowpass_check_state
        self.lowpass_text.disabled = lowpass_cut_state
        self.highpass_checkbox.disabled = highpass_cut_state
        self.highpass_text.disabled = highpass_cut_state
        self.list_box.disabled = list_box_state

    def about(self, event):
        popup = Popup(title="About UVicMUSE", content=Label(text='Developed and Designed by \nBardia'
                                                                 'Barabadi\n\nKrigolson Lab '
                                                                 '(Theoretical and \nApplied Neuroscience '
                                                                 'Laboratory)\n\n '
                                                                 'University of Victoria, Canada, 2020.'
                                                            ), size_hint=(None, None), size=(300, 200))
        popup.open()

    def reset(self, event):
        if self.did_connect:
            self.disconnect(event)
        else:
            self.muses = []
            self.muse = []
            self.host_address = 'localhost'
            self.backend = Backend()
            self.current_muse_id = 0
            self.did_connect = False
            self.connect_button.text = "Connect"
            self.connect_button.disabled = True
            self.list_box.values = []
            self.list_box.text = self.press_search_txt

    # Function for starting Data Stream
    def stream(self, event):

        if self.backend.is_udp_streaming:
            self.backend.udp_stop_btn_callback()
            self.status_label.text = "Data stream has been stopped"
            self.stream_button.text = "Start Streaming"

            self.lowpass_checkbox.disabled = False
            self.highpass_checkbox.disabled = False
            self.notch_checkbox.disabled = False
            self.highpass_text.disabled = False
            self.lowpass_text.disabled = False
            # self.LSL_checkbox.disabled = False

            return

        self.button_state(True, False, True, True, True, True, True)

        self.backend.udp_stream_btn_callback(
            use_low_pass=self.lowpass_checkbox.active,
            use_high_pass=self.highpass_checkbox.active,
            low_pass_cutoff=float(self.get_lowpass_cutoff()),
            high_pass_cutoff=float(self.get_highpass_cutoff()),
            use_notch=self.get_notch_checkbox, )
        if self.backend.is_udp_streaming:
            self.status_label.text = "Streaming Data"
            self.stream_button.text = "Stop Streaming"
        else:
            self.status_label.text = "Unsuccessful streaming attempt"

    # Function for stopping Data Stream
    def stop_stream(self, event):
        self.button_state(True, False, False, False, False, False, True)
        self.backend.udp_stop_btn_callback()
        self.status_label.text = "Data stream has been stopped"

    def get_lowpass_cutoff(self):
        return float(self.lowpass_text.text)

    def get_highpass_cutoff(self):
        return float(self.highpass_text.text)

    def get_notch_checkbox(self):
        if self.notch_checkbox.active:
            return True
        return False

    def get_host_entry(self):
        return str(self.host_text.text)

    # Function for Connecting to selected Muse in Dropdown list
    def connect(self, event):

        if self.did_connect:
            self.disconnect(event)
            return

        try:
            self.button_state(True, False, False, False, False, False, True)

            self.current_muse_id = self.list_box.values.index(self.list_box.text)

            self.backend.connect_btn_callback(self.current_muse_id)
            self.did_connect = self.backend.is_connected()
            if self.did_connect:
                self.status_label.text = "Successfully connected to " + str(
                    self.muses[self.current_muse_id].name + ", select filters to stream data with")
                self.connected_address = self.muses[self.current_muse_id].address
                self.connect_button.text = "Disconnect"
            else:
                self.status_label.text = "Unsuccessful connection attempt, make sure Muse is turned on"
        except:
            self.status_label.text = "Please select a Muse from the dropdown menu before connecting"
            self.button_state(False,
                              True, True, True,
                              True, True, False)

    # Function for Disconnecting from connected Muse
    def disconnect(self, event):
        self.button_state(False, True, False,
                          False, False, False, False)

        self.backend.disconnect_btn_callback()
        self.status_label.text = "Disconnected from " + str(self.muses[self.current_muse_id].name)
        self.did_connect = False
        self.connect_button.text = "Connect"
        self.connect_button.disabled = True
        self.list_box.values = []
        self.list_box.text = self.press_search_txt

    # Function for Searching for nearby Muses
    def search(self, event):

        if self.did_connect:
            self.disconnect(event)

        try:
            self.button_state(False,
                              True, True, True,
                              True, True, False)

            self.connect_button.disabled = False

            self.muses, succeed = self.backend.refresh_btn_callback(event)
            self.vals = []
            for i in range(len(self.muses)):
                self.vals.append(self.muses[i].name + ", Address: " + str(self.muses[i].address))

            if (len(self.muses)) == 0:
                self.status_label.text = "No nearby devices were found, please try again                               "
            if (len(self.muses)) == 1:
                self.status_label.text = "1 device was found, please choose sensors to connect with"
            if (len(self.muses)) > 1:
                self.status_label.text = str(
                    len(self.muses)) + " Devices were found, please choose sensors to connect with"

            self.list_box.values = self.vals

            self.list_box.text = str(len(self.muses)) + " Devices were found, Press to choose"
        except:
            self.status_label.text = "No BLE Module Found"


class MuseApp(App):
    title = "uvicMuse"

    def build(self):
        resource_add_path(resource_path('.'))

        return UVicMuse()


def runGUI():
    loop = asyncio.get_event_loop()
    loop.run_until_complete(App.async_run(MuseApp(), async_lib='asyncio'))
    loop.close()
    # MuseApp().run()


# from .constants import *
from uvicmuse.constants import *

import bitstring
from bleak import BleakClient
import json
from bleak.exc import BleakError
import numpy as np
from time import time
import asyncio


def ppg_error():
    print("PPG is only available on MUSE v2")
    # T0D0: a proper error routine


class MuseBLE(object):

    def __init__(self, client,
                 callback_control=None,  # Will be called with a dictionary containing control info
                 callback_eeg=None,
                 callback_ppg=None,
                 callback_acc=None,
                 callback_gyro=None,
                 callback_tele=None
                 ):

        self.loop = asyncio.new_event_loop()
        self.client = BleakClient(client.address, loop=self.loop)

        self.last_tm_ppg = 0
        self.last_tm_eeg = 0
        self.callback_control = callback_control
        self.callback_eeg = callback_eeg
        self.callback_ppg = callback_ppg
        self.callback_acc = callback_acc
        self.callback_gyro = callback_gyro
        self.callback_tele = callback_tele

        self._current_control_seq = ""

    async def connect(self, timeout=20):
        # Tries to connect to the given MUSE (given the address in the constructor), Returns success

        success = False
        for retry in range(int(timeout / 4) + 1):
            try:
                await self.client.connect(timeout=timeout)
                success = True
            except BleakError:
                success = False
            if success:
                break
        if not success:
            return False
        if self.callback_control is not None:
            await self._subscribe_control()
        if self.callback_eeg is not None:
            await self._subscribe_eeg()
        if self.callback_ppg is not None and len(self.client.services.characteristics) > 12:
            await self._subscribe_ppg()
        if self.callback_acc is not None:
            await self._subscribe_acc()
        if self.callback_tele is not None:
            await self._subscribe_telemetry()
        if self.callback_gyro is not None:
            await self._subscribe_gyro()

        return True

    def disconnect_(self):
        self.loop.run_until_complete(self.disconnect())

    async def disconnect(self):
        # Disconnects from the MUSE, returns success
        try:
            await self.client.disconnect()
            success = True
        except BleakError:
            success = False
        return success

    async def ask_control(self):
        """Send a message to Muse to ask for the control status.
                The message received is a dict with the following keys:
                "hn": device name
                "sn": serial number
                "ma": MAC address
                "id":
                "bp": battery percentage
                "ts":
                "ps": preset selected
                "rc": return status, if 0 is OK
        """
        await self._write_command([0x02, 0x73, 0x0a])

    async def ask_device_info(self):
        """Send a message to Muse to ask for the device info.
                The message received is a dict with the following keys:
                "ap":
                "sp":
                "tp": firmware type, e.g: "consumer"
                "hw": hardware version?
                "bn": build number?
                "fw": firmware version?
                "bl":
                "pv": protocol version?
                "rc": return status, if 0 is OK
        """
        await self._write_command([0x03, 0x76, 0x31, 0x0a])

    async def ask_reset(self):
        # The message received would be {"rc": return status}, 0 is OK
        await self._write_command([0x03, 0x2a, 0x31, 0x0a])

    async def start(self):
        self._init_timestamp_correction()
        self._init_sample_eeg()
        self._init_sample_ppg()
        self._init_sample_control()
        self.last_tm_eeg = 0
        self.last_tm_ppg = 0
        await self.resume()

    async def resume(self):
        await self._write_command([0x02, 0x64, 0x0a])

    async def stop(self):
        await self._write_command([0x02, 0x68, 0x0a])

    async def keep_alive(self):
        await self._write_command([0x02, 0x6b, 0x0a])

    async def select_preset(self, preset):
        # Setting preset for headband configuration
        if preset == 20:
            await self._write_command([0x04, 0x70, 0x32, 0x30, 0x0a])
        elif preset == 22:
            await self._write_command([0x04, 0x70, 0x32, 0x32, 0x0a])
        elif preset == 23:
            await self._write_command([0x04, 0x70, 0x32, 0x33, 0x0a])
        else:
            await self._write_command([0x04, 0x70, 0x32, 0x31, 0x0a])

    async def _subscribe_to_uuid(self, uuid, callback):
        await self.client.start_notify(uuid, callback)

    async def _write_command(self, command):
        await self.client.write_gatt_char(MUSE_CONTROL_HANDLE, bytearray(command), False)

    # _subscribe functions

    async def _subscribe_control(self):
        await self._subscribe_to_uuid(MUSE_CONTROL_HANDLE, self._handle_control)

    async def _subscribe_eeg(self):
        await self._subscribe_to_uuid(uuid=MUSE_GATT_ATTR_TP9, callback=self._handle_eeg)
        await self._subscribe_to_uuid(uuid=MUSE_GATT_ATTR_AF7, callback=self._handle_eeg)
        await self._subscribe_to_uuid(uuid=MUSE_GATT_ATTR_AF8, callback=self._handle_eeg)
        await self._subscribe_to_uuid(uuid=MUSE_GATT_ATTR_TP10, callback=self._handle_eeg)
        await self._subscribe_to_uuid(uuid=MUSE_GATT_ATTR_RIGHTAUX, callback=self._handle_eeg)

    async def _subscribe_ppg(self):
        try:
            await self._subscribe_to_uuid(uuid=MUSE_GATT_ATTR_PPG1, callback=self._handle_ppg)
            await self._subscribe_to_uuid(uuid=MUSE_GATT_ATTR_PPG2, callback=self._handle_ppg)
            await self._subscribe_to_uuid(uuid=MUSE_GATT_ATTR_PPG3, callback=self._handle_ppg)
        except BleakError:
            raise ppg_error

    async def _subscribe_gyro(self):
        await self._subscribe_to_uuid(uuid=MUSE_GATT_ATTR_GYRO, callback=self._handle_gyro)

    async def _subscribe_acc(self):
        await self._subscribe_to_uuid(uuid=MUSE_GATT_ATTR_ACCELEROMETER, callback=self._handle_acc)

    async def _subscribe_telemetry(self):
        # Handles telemetry (Battery, temp, ...) incoming data
        await self._subscribe_to_uuid(uuid=MUSE_GATT_ATTR_TELEMETRY, callback=self._handle_tele)

    # Unpack functions
    @staticmethod
    def _unpack_eeg_channel(packet):
        # Each packet is encoded with a 16bit timestamp followed by 12 time samples with a 12 bit resolution.
        aa = bitstring.Bits(bytes=packet)
        pattern = "uint:16,uint:12,uint:12,uint:12,uint:12,uint:12,uint:12, \
                          uint:12,uint:12,uint:12,uint:12,uint:12,uint:12"
        res = aa.unpack(pattern)
        packet_index = res[0]
        data = res[1:]
        # 12 bits on a 2 mVpp range
        data = 0.48828125 * (np.array(data) - 2048)
        return packet_index, data

    @staticmethod
    def _unpack_ppg_channel(packet):
        """Decode data packet of one PPG channel.
        Each packet is encoded with a 16bit timestamp followed by 3
        samples with an x bit resolution.
        """

        aa = bitstring.Bits(bytes=packet)
        pattern = "uint:16,uint:24,uint:24,uint:24,uint:24,uint:24,uint:24"
        res = aa.unpack(pattern)
        packet_index = res[0]
        data = res[1:]

        return packet_index, data

    @staticmethod
    def _unpack_imu_channel(packet, scale=1):
        """Decode data packet of the accelerometer and gyro (imu) channels.
                Each packet is encoded with a 16bit timestamp followed by 9 samples
                with a 16 bit resolution.
                """
        bit_decoder = bitstring.Bits(bytes=packet)
        pattern = "uint:16,int:16,int:16,int:16,int:16, \
                           int:16,int:16,int:16,int:16,int:16"
        data = bit_decoder.unpack(pattern)

        packet_index = data[0]

        samples = np.array(data[1:]).reshape((3, 3), order='F') * scale

        return packet_index, samples

    # Initialize samples

    def _init_sample_control(self):
        self._current_control_seq = ""

    def _init_sample_eeg(self):
        self.timestamps_eeg = np.zeros(5)
        self.data_eeg = np.zeros((5, 12))

    def _init_sample_ppg(self):
        self.timestamps_ppg = np.zeros(3)
        self.data_ppg = np.zeros((3, 6))

    # Timestamp Correction
    def _init_timestamp_correction(self):
        """Init IRLS params"""
        # initial params for the timestamp correction
        # the time it started + the inverse of sampling rate
        self.sample_index_eeg = 0
        self.sample_index_ppg = 0
        self.reg_params = np.array(
            [time(), 1. / MUSE_SAMPLING_EEG_RATE])
        self.reg_ppg_sample_rate = np.array(
            [time(), 1. / MUSE_SAMPLING_PPG_RATE])

    # Handlers
    def _handle_control(self, sender, packet):
        # Handles the message coming from the control sequence
        assert sender == 13, "_handle_control is receiving a message " \
                             "with a different UUID " + str(sender)

        bit_decoder = bitstring.Bits(bytes=packet)
        pattern = "uint:8,uint:8,uint:8,uint:8,uint:8,uint:8,uint:8,uint:8,uint:8,uint:8, \
                                uint:8,uint:8,uint:8,uint:8,uint:8,uint:8,uint:8,uint:8,uint:8,uint:8"
        chars = bit_decoder.unpack(pattern)

        # Length of the string
        n_incoming = chars[0]

        # Parse as chars, only useful bytes
        incoming_message = "".join(map(chr, chars[1:]))[:n_incoming]

        self._current_control_seq += incoming_message

        if incoming_message[-1] == '}':  # end of the current message
            self.callback_control(json.loads(self._current_control_seq))  # Call the callback
            self._init_sample_control()

    def _handle_eeg(self, sender, packet):
        index = int((sender - 32) / 3)
        timestamp = time()
        tm, d = self._unpack_eeg_channel(packet)

        if self.last_tm_eeg == 0:
            self.last_tm_eeg = tm - 1

        self.data_eeg[index] = d
        self.timestamps_eeg[index] = timestamp

        # Check if this is the last data in the sequence
        if index == 1:
            if tm != self.last_tm_eeg + 1:
                pass
                # print("Missing sample %d : %d" % (tm, self.last_tm_eeg))
            self.last_tm_eeg = tm

            idxs = np.arange(0, 12) + self.sample_index_eeg
            self.sample_index_eeg += 12

            # timestamps are extrapolated backwards based on sampling rate and current time
            timestamps = self.reg_params[1] * idxs + self.reg_params[0]

            # push data
            self.callback_eeg(self.data_eeg, timestamps)

            # save last timestamp for disconnection timer
            self.last_timestamp = timestamps[-1]

            # reset sample
            self._init_sample_eeg()

    def _handle_ppg(self, sender, data):
        """Callback for receiving a sample.
        samples are received in this order : 56, 59, 62
        wait until we get x and call the data callback
        """

        handle = sender - 1
        timestamp = time()
        index = int((handle - 56) / 3)
        tm, d = self._unpack_ppg_channel(data)

        if self.last_tm_ppg == 0:
            self.last_tm_ppg = tm - 1

        self.data_ppg[index] = d
        self.timestamps_ppg[index] = timestamp
        # last data received
        if handle == 62:
            if tm != self.last_tm_ppg + 1:
                pass
                # print("missing sample %d : %d" % (tm, self.last_tm_ppg))
            self.last_tm_ppg = tm

            # calculate index of time samples
            indices = np.arange(0, LSL_PPG_CHUNK) + self.sample_index_ppg
            self.sample_index_ppg += LSL_PPG_CHUNK

            # timestamps are extrapolated backwards based on sampling rate and current time
            timestamps = self.reg_ppg_sample_rate[1] * indices + self.reg_ppg_sample_rate[0]

            # save last timestamp for disconnection timer
            self.last_timestamp = timestamps[-1]

            # push data
            if self.callback_ppg:
                self.callback_ppg(self.data_ppg, timestamps)

            # reset sample
            self._init_sample_ppg()

    def _handle_acc(self, sender, packet):
        """Handle incoming accelerometer data.
        sampling rate: ~17 x second (3 samples in each message, roughly 50Hz)"""
        if sender != 22:  # handle 0x17
            print("ERROR ACC" + str(sender))
            return

        timestamps = [time()] * 3

        # save last timestamp for disconnection timer
        self.last_timestamp = timestamps[-1]

        packet_index, samples = self._unpack_imu_channel(
            packet, scale=MUSE_ACCELEROMETER_SCALE_FACTOR)

        self.callback_acc(samples, timestamps)

    def _handle_gyro(self, sender, packet):
        """Handle incoming gyroscope data.
        sampling rate: ~17 x second (3 samples in each message, roughly 50Hz)"""
        if sender != 19:  # handle 0x14
            print("ERROR GYRO" + str(sender))
            return
        timestamps = [time()] * 3

        # save last timestamp for disconnection timer
        self.last_timestamp = timestamps[-1]

        packet_index, samples = self._unpack_imu_channel(
            packet, scale=MUSE_GYRO_SCALE_FACTOR)

        self.callback_gyro(samples, timestamps)

    def _handle_tele(self, sender, packet):
        """Handle the telemetry (battery, temperature and stuff) incoming data
        """
        if sender != 26:  # handle 0x1a
            print("ERROR TELE" + str(sender))
            return
        timestamp = time()

        bit_decoder = bitstring.Bits(bytes=packet)
        pattern = "uint:16,uint:16,uint:16,uint:16,uint:16"  # The rest is 0 padding
        data = bit_decoder.unpack(pattern)

        battery = data[1] / 512
        fuel_gauge = data[2] * 2.2
        adc_volt = data[3]
        temperature = data[4]

        self.callback_tele(
            timestamp, battery, fuel_gauge, adc_volt, temperature)

    def command_callback(dc):
        print("Received command callback completely: ")
        print(dc)



from uvicmuse.constants import *
from uvicmuse.MuseBLE import *
from uvicmuse.MuseFinder import *
from uvicmuse.helper import *

import asyncio
from functools import partial

from pylsl import StreamInfo, StreamOutlet


class MuseWrapper:

    def __init__(self,
                 loop,
                 target_name=None,  # Should be a string containing four digits/letters
                 timeout=10,
                 max_buff_len=2 * 256  # Set this to be the maximum number of samples to keep
                 ):
        self.all_muses = []
        self.target_name = target_name
        self.timeout = timeout
        self.muse = None
        self.target_muse = None

        self.eeg_buff = []
        self.ppg_buff = []
        self.acc_buff = []
        self.gyro_buff = []
        self.max_buff_len = max_buff_len
        self.loop = loop

    def disconnect(self):
        self.loop.run_until_complete(self.muse.disconnect())

    def search_and_connect(self):
        mf = MuseFinder(add_muse_to_list_callback=None)
        loop = asyncio.get_event_loop()
        loop.run_until_complete(mf.search_for_muses(timeout=self.timeout))
        self.all_muses = mf.get_muses()

        success = False

        if len(self.all_muses) < 1:
            print("No MUSEs found, please try again or increase the timeout")
            return success

        if self.target_name is None:
            if len(self.all_muses) == 1:
                print("No target MUSE specified, one device found. Connecting to " + str(self.all_muses[0].name))
                self.target_muse = self.all_muses[0]
            else:
                print("No target MUSE specified, more than one MUSE is in range. Please specify the target.")
                return success

        else:
            found = False
            for d in self.all_muses:
                if self.target_name in d.name:
                    print("Target MUSE-" + self.target_name + " found. Attempting to connect...")
                    self.target_muse = d
                    found = True
            if found is False:
                print("Couldn't find target, + " + str(self.target_name) + ". Please specify the target.")
                return False

        # Connecting

        eeg_outlet = self._get_eeg_outlet()
        ppg_outlet = self._get_ppg_outlet()
        acc_outlet = self._get_acc_outlet()
        gyro_outlet = self._get_gyro_outlet()

        push_eeg = partial(self._push, outlet=eeg_outlet, offset=EEG_PORT_OFFSET)
        push_ppg = partial(self._push, outlet=ppg_outlet, offset=PPG_PORT_OFFSET)
        push_acc = partial(self._push, outlet=acc_outlet, offset=ACC_PORT_OFFSET)
        push_gyro = partial(self._push, outlet=gyro_outlet, offset=GYRO_PORT_OFFSET)
        self.muse = MuseBLE(client=self.target_muse, callback_control=self._command_callback,
                            callback_acc=push_acc, callback_eeg=push_eeg, callback_gyro=push_gyro,
                            callback_ppg=push_ppg)

        loop.run_until_complete(self.muse.connect(timeout=30))

        print("connection was successful")
        asyncio.run(asyncio.sleep(2))
        success = True
        return success

    def pull_eeg(self):
        self.loop.run_until_complete(self.muse.start())
        to_return = self.eeg_buff
        self.eeg_buff = []
        return to_return

    def pull_ppg(self):
        self.loop.run_until_complete(self.muse.start())
        to_return = self.ppg_buff
        self.ppg_buff = []
        return to_return

    def pull_acc(self):
        self.loop.run_until_complete(self.muse.start())
        to_return = self.acc_buff
        self.acc_buff = []
        return to_return

    def pull_gyro(self):
        self.loop.run_until_complete(self.muse.start())
        to_return = self.gyro_buff
        self.gyro_buff = []
        return to_return

    def reset_buffers(self):
        self.eeg_buff = []
        self.ppg_buff = []
        self.acc_buff = []
        self.gyro_buff = []

    def get_muse_address(self):
        return self.muse.address

    def get_muse(self):
        return self.muse

    def _get_eeg_outlet(self):
        # Connecting to MUSE
        eeg_info = StreamInfo('Muse', 'EEG', MUSE_NB_EEG_CHANNELS, MUSE_SAMPLING_EEG_RATE, 'float32',
                              'Muse%s' % self.get_muse_address())
        eeg_info.desc().append_child_value("manufacturer", "Muse")
        eeg_channels = eeg_info.desc().append_child("channels")

        for c in ['TP9', 'AF7', 'AF8', 'TP10', 'Right AUX']:
            eeg_channels.append_child("channel") \
                .append_child_value("label", c) \
                .append_child_value("unit", "microvolts") \
                .append_child_value("type", "EEG")

        eeg_outlet = StreamOutlet(eeg_info, LSL_EEG_CHUNK)

        return eeg_outlet

    def _get_ppg_outlet(self):
        # Connecting to MUSE
        ppg_info = StreamInfo('Muse', 'PPG', MUSE_NB_PPG_CHANNELS, MUSE_SAMPLING_PPG_RATE,
                              'float32', 'Muse%s' % self.get_muse_address())
        ppg_info.desc().append_child_value("manufacturer", "Muse")

        ppg_outlet = StreamOutlet(ppg_info, LSL_PPG_CHUNK)

        return ppg_outlet

    def _get_acc_outlet(self):
        acc_info = StreamInfo('Muse', 'ACC', MUSE_NB_ACC_CHANNELS, MUSE_SAMPLING_ACC_RATE,
                              'float32', 'Muse%s' % self.get_muse_address())
        acc_info.desc().append_child_value("manufacturer", "Muse")

        acc_outlet = StreamOutlet(acc_info, LSL_ACC_CHUNK)

        return acc_outlet

    def _get_gyro_outlet(self):
        gyro_info = StreamInfo('Muse', 'GYRO', MUSE_NB_GYRO_CHANNELS, MUSE_SAMPLING_GYRO_RATE,
                               'float32', 'Muse%s' % self.get_muse_address())
        gyro_info.desc().append_child_value("manufacturer", "Muse")

        gyro_outlet = StreamOutlet(gyro_info, LSL_GYRO_CHUNK)

        return gyro_outlet

    def _push(self, data, timestamps, outlet, offset=0):
        for ii in range(data.shape[1]):
            if not is_data_valid(data[:, ii], timestamps[ii]):
                continue
            outlet.push_sample(data[:, ii], timestamps[ii])
            if offset == EEG_PORT_OFFSET:
                to_append_eeg_data = [data[0, ii], data[1, ii], data[2, ii], data[3, ii], data[4, ii],
                                      (timestamps[ii])]
                if len(self.eeg_buff) >= self.max_buff_len:
                    self.eeg_buff.pop(0)
                self.eeg_buff.append(to_append_eeg_data)

            elif offset == PPG_PORT_OFFSET:
                to_append_ppg_data = [data[0, ii], data[1, ii], data[2, ii],
                                      (timestamps[ii])]
                if len(self.ppg_buff) >= self.max_buff_len:
                    self.ppg_buff.pop(0)
                self.ppg_buff.append(to_append_ppg_data)

            elif offset == ACC_PORT_OFFSET:
                to_append_acc_data = [data[0, ii], data[1, ii], data[2, ii],
                                      (timestamps[ii])]
                if len(self.acc_buff) >= self.max_buff_len:
                    self.acc_buff.pop(0)
                self.acc_buff.append(to_append_acc_data)

            elif offset == GYRO_PORT_OFFSET:
                to_append_gyro_data = [data[0, ii], data[1, ii], data[2, ii],
                                       (timestamps[ii])]
                if len(self.gyro_buff) >= self.max_buff_len:
                    self.gyro_buff.pop(0)
                self.gyro_buff.append(to_append_gyro_data)

    @staticmethod
    def _command_callback(a):
        pass
        # print("Received command callback: ")
        # print(a)



# -*- coding: utf-8 -*-

__all__ = ('start', 'sleep', 'event', )

import types
from functools import partial
from collections import namedtuple
from kivy.clock import Clock

CallbackParameter = namedtuple('CallbackParameter', ('args', 'kwargs', ))


def start(coro):

    def step(*args, **kwargs):
        try:
            coro.send(CallbackParameter(args, kwargs))(step)
        except StopIteration:
            pass

    try:
        coro.send(None)(step)
    except StopIteration:
        pass


@types.coroutine
def sleep(duration):
    # The partial() here looks meaningless. But this is needed in order
    # to avoid weak reference
    param = yield lambda step_coro: Clock.schedule_once(
        partial(step_coro), duration)
    return param.args[0]


class event:

    def __init__(self, ed, name):
        self.bind_id = None
        self.ed = ed
        self.name = name

    def bind(self, step_coro):
        self.bind_id = bind_id = self.ed.fbind(self.name, self.callback)
        assert bind_id > 0  # check if binding succeeded
        self.step_coro = step_coro

    def callback(self, *args, **kwargs):
        self.parameter = CallbackParameter(args, kwargs)
        ed = self.ed
        ed.unbind_uid(self.name, self.bind_id)
        self.step_coro()

    def __await__(self):
        yield self.bind
        return self.parameter


from bleak import discover

# from .constants import *
from uvicmuse.constants import *

class MuseFinder(object):

    def __init__(self,
                 add_muse_to_list_callback=None,  # Used to call a UI function when a new muse is found
                 ):
        self.muses = []
        self.add_muse_to_list_callback = add_muse_to_list_callback

    async def search_for_muses(self, timeout=1):
        self.muses = []
        for i in range(int(timeout / 2) + 1):
            # search for muses timeout/2 times, 2 seconds each time
            devices = await discover(timeout=2)
            for d in devices:
                if MUSE_NAME in d.name and not self._is_already_found(newly_found_muse=d):
                    self.muses.append(d)
                    if self.add_muse_to_list_callback is not None:
                        self.add_muse_to_list_callback(d)

    def _is_already_found(self, newly_found_muse):
        for old_muse in self.muses:
            if newly_found_muse.name == old_muse.name:
                return True
        return False

    def get_muses(self):
        return self.muses
